__version__ = "0.1.0"
__author__ = "Lightning-AI et al."
__author_email__ = "luca@lightning.ai"
__license__ = "Apache-2.0"
__copyright__ = f"Copyright (c) 2022-2023, {__author__}."
__homepage__ = "https://github.com/Lightning-AI/lightning-GPT"
__docs__ = "GPT training in Lightning"
__docs_url__ = "https://github.com/Lightning-AI/lightning-GPT"
__long_doc__ = """
minGPT training and benchmarking with Lightning
"""

__all__ = [
    "__author__",
    "__author_email__",
    "__copyright__",
    "__docs__",
    "__docs_url__",
    "__homepage__",
    "__license__",
    "__version__",
]
